from pprint import pformat


class ConfigNode(object):
    """
    A node in the configuration store.
    These are obtained by using the methods below on :class:`~configurator.Config`
    objects or any :class:`ConfigNode` objects returned by them.
    """

    __slots__ = ('data',)

    def __init__(self, data=None):
        if data is None:
            data = {}
        #: The underlying python object for this node, often a :class:`dict`
        #: or a :class:`list`. While this is part of the public API, care should
        #: be taken when using it, particularly if you choose to modify it.
        self.data = data

    @classmethod
    def _wrap(cls, value):
        if isinstance(value, (dict, list)):
            value = ConfigNode(value)
        return value

    def _get(self, name):
        try:
            value = self.data[name]
        except TypeError:
            raise KeyError(name)
        return self._wrap(value)

    def __getattr__(self, name):
        """
        Obtain a child of this node by attribute access. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        try:
            return self._get(name)
        except KeyError:
            raise AttributeError(name)

    def __getitem__(self, name):
        """
        Obtain a child of this node by item access. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        return self._get(name)

    def get(self, name, default=None):
        """
        Obtain a child of this node by access like :meth:`dict.get`. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        try:
            return self._get(name)
        except KeyError:
            return default

    def items(self):
        """
        Obtain children of this node by access like :meth:`dict.items`. If the child
        is a :class:`dict` or :class:`list`, a :class:`ConfigNode` for it will
        be returned, otherwise the value itself will be returned.
        """
        for key, value in sorted(self.data.items()):
            yield key, self._wrap(value)

    def __iter__(self):
        """
        Obtain children of this node by either :class:`dict` or :class:`list`
        iteration, depending on the type of the underlying :attr:`data`.
        If the child is a :class:`dict` or :class:`list`, a :class:`ConfigNode`
        for it will be returned, otherwise the value itself will be returned.
        """
        for item in self.data:
            yield self._wrap(item)

    def __repr__(self):
        cls = type(self)
        pretty = pformat(self.data, width=70)
        if '\n' in pretty:
            pretty = '\n'+pretty+'\n'
        return '{}.{}({})'.format(
            cls.__module__,
            cls.__name__,
            pretty
        )



from pkg_resources import iter_entry_points


class ParseError(Exception):
    """
    The exception raised when an appropriate parser cannot be found for a config
    stream.
    """


class Parsers(dict):
    
    @classmethod
    def from_entrypoints(cls):
        parsers = cls()
        for entrypoint in iter_entry_points(group='configurator.parser'):
            try:
                parsers[entrypoint.name] = entrypoint.load()
            except ImportError:
                # a package may present entry points based on soft dependencies,
                # which may not be available.
                pass
        return parsers

    def get(self, extension):
        try:
            return self[extension]
        except KeyError:
            raise ParseError('No parser found for {!r}'.format(extension))



from ast import literal_eval
from tempfile import NamedTemporaryFile

import pytest

from configurator import Config, default_mergers
from io import StringIO
from configurator.parsers import ParseError
from configurator.mapping import source, target, convert
from testfixtures import compare, ShouldRaise, TempDirectory, Replace

from .compat import type_error


def python_literal(stream):
    return literal_eval(stream.read())


class TestInstantiation(object):

    def test_empty(self):
        config = Config()
        compare(config.data, expected={})

    def test_dict(self):
        config = Config(dict(x=1))
        compare(config.x, expected=1)

    def test_list(self):
        config = Config([1, 2])
        compare(config[0], expected=1)
        compare(config[1], expected=2)
        compare(list(config), expected=[1, 2])

    def test_int(self):
        # not very useful...
        config = Config(1)
        compare(config.data, expected=1)

    def test_text_string_parser(self):
        config = Config.from_text('{"foo": "bar"}', 'json')
        compare(config.data, expected={'foo': 'bar'})

    def test_bytes_string_parser(self):
        config = Config.from_text(b'{"foo": "bar"}', 'json')
        compare(config.data, expected={'foo': 'bar'})

    def test_text_callable_parser(self):
        config = Config.from_text("{'foo': 'bar'}", python_literal)
        compare(config.data, expected={'foo': 'bar'})

    def test_text_missing_parser(self):
        with ShouldRaise(ParseError("No parser found for 'lolwut'")):
            Config.from_text("{'foo': 'bar'}", 'lolwut')

    def test_path_guess_parser(self):
        with NamedTemporaryFile(suffix='.json') as source:
            source.write(b'{"x": 1}')
            source.flush()
            config = Config.from_path(source.name)
        compare(config.x, expected=1)

    def test_path_guess_parser_no_extension(self):
        with TempDirectory() as dir:
            path = dir.write('nope', b'{"x": 1}')
            with ShouldRaise(ParseError("No parser found for None")):
                Config.from_path(path)

    def test_path_guess_parser_bad_extension(self):
        with NamedTemporaryFile(suffix='.nope') as source:
            with ShouldRaise(ParseError("No parser found for 'nope'")):
                Config.from_path(source.name)

    def test_path_explicit_string_parser(self):
        with NamedTemporaryFile() as source:
            source.write(b'{"x": 1}')
            source.flush()
            config = Config.from_path(source.name, 'json')
        compare(config.x, expected=1)

    def test_path_explicit_callable_parser(self):
        with NamedTemporaryFile() as source:
            source.write(b'{"x": 1}')
            source.flush()
            config = Config.from_path(source.name, python_literal)
        compare(config.x, expected=1)

    def test_path_with_encoding(self):
        with NamedTemporaryFile() as source:
            source.write(b'{"x": "\xa3"}')
            source.flush()
            config = Config.from_path(source.name, 'json', encoding='latin-1')
        compare(config.x, expected=u'\xa3')

    def test_stream_with_name_guess_parser(self):
        with NamedTemporaryFile(suffix='.json') as source:
            source.write(b'{"x": 1}')
            source.flush()
            source.seek(0)
            config = Config.from_stream(source)
        compare(config.x, expected=1)

    def test_stream_no_name_no_parser(self):
        source = StringIO(u'{"x": 1}')
        with ShouldRaise(ParseError("No parser found for None")):
            Config.from_stream(source)

    def test_stream_string_parser(self):
        source = StringIO(u'{"x": 1}')
        config = Config.from_stream(source, 'json')
        compare(config.x, expected=1)

    def test_stream_callable_parser(self):
        source = StringIO(u'{"x": 1}')
        config = Config.from_stream(source, python_literal)
        compare(config.x, expected=1)

    @pytest.fixture()
    def env(self):
        env = {}
        with Replace('os.environ', env):
            yield env

    def test_from_env_single_prefix(self, env):
        env['FOO_BAR'] = 'one'
        env['FOO_BAZ'] = 'two'
        config = Config.from_env(prefix='FOO_')
        compare(config.data, expected={'bar': 'one', 'baz': 'two'})

    def test_from_env_multiple_prefix(self, env):
        env['FOO_BAR'] = 'one'
        env['BOB_BAR'] = 'two'
        config = Config.from_env({'FOO_': 'foo', 'BOB_': 'bob'})
        compare(config.data, expected={
            'foo': {'bar': 'one'},
            'bob': {'bar': 'two'}
        })

    def test_from_env_target_dotted_string(self, env):
        env['FOO_BAR'] = 'one'
        env['FOO_BAZ'] = 'two'
        config = Config.from_env(prefix={'FOO_': 'a.b.c'})
        compare(config.data, expected={
            'a': {'b': {'c': {'bar': 'one', 'baz': 'two'}}}
        })

    def test_from_env_target_path(self, env):
        env['FOO_BAR'] = 'one'
        env['FOO_BAZ'] = 'two'
        config = Config.from_env(prefix={'FOO_': target['a']})
        compare(config.data, expected={
            'a': {'bar': 'one', 'baz': 'two'}
        })

    def test_from_env_type_suffix(self, env):
        env['FOO_BAR'] = '1'
        env['FOO_BAZ'] = '2'
        config = Config.from_env(prefix='FOO_', types={'_BAZ': int})
        compare(config.data, expected={'bar': '1', 'baz': 2})

    def test_from_env_value_is_empty_string(self, env):
        env['FOO_BAR'] = ''
        env['FOO_BAZ'] = ''
        config = Config.from_env(prefix='FOO_', types={'_BAZ': int})
        compare(config.data, expected={})


class TestPushPop(object):

    def test_push_pop(self):
        config = Config({'x': 1, 'y': 2})
        compare(config.x, expected=1)
        compare(config.y, expected=2)
        config.push(Config({'x': 3}))
        compare(config.x, expected=3)
        compare(config.y, expected=2)
        config.pop()
        compare(config.x, expected=1)
        compare(config.y, expected=2)

    def test_push_empty(self):
        config = Config({'x': 1, 'y': 2})
        compare(config.x, expected=1)
        compare(config.y, expected=2)
        config.push(Config({'x': 3}), empty=True)
        compare(config.x, expected=3)
        with ShouldRaise(AttributeError('y')):
            config.y
        config.pop()
        compare(config.x, expected=1)
        compare(config.y, expected=2)

    def test_push_non_config(self):
        config = Config({'x': 1})
        compare(config.x, expected=1)
        config.push({'x': 2})
        compare(config.x, expected=2)

    def test_pop_without_push(self):
        config = Config({'x': 1, 'y': 2})
        with ShouldRaise(IndexError('pop from empty list')):
            config.pop()

    def test_context_manager_push(self):
        config = Config({'x': 1, 'y': 2})
        compare(config.x, expected=1)
        compare(config.y, expected=2)
        with config.push(Config({'x': 3})):
            compare(config.x, expected=3)
            compare(config.y, expected=2)
        compare(config.x, expected=1)
        compare(config.y, expected=2)

    def test_context_manager_push_pathological(self):
        config = Config({'x': 1, 'y': 2})
        compare(config.x, expected=1)
        compare(config.y, expected=2)
        with config.push():
            config.data['a'] = 5
            config.push({'x': 3})
            config.push({'z': 4})
            compare(config.a, expected=5)
            compare(config.x, expected=3)
            compare(config.y, expected=2)
            compare(config.z, expected=4)
        compare(config.x, expected=1)
        compare(config.y, expected=2)
        with ShouldRaise(AttributeError('a')):
            config.a
        with ShouldRaise(AttributeError('z')):
            config.z

    def test_context_manager_push_deep(self):
        config = Config({'x': {'y': 'z'}})
        with config.push():
            config.data['x']['y'] = 'a'
            compare(config.x.y, expected='a')
        compare(config.x.y, expected='z')


class TestNodeBehaviour(object):

    def test_dict_access(self):
        config = Config({'foo': 'bar'})
        compare(config['foo'], expected='bar')

    def test_attr_access(self):
        config = Config({'foo': 'bar'})
        compare(config.foo, expected='bar')


class TestMergeTests(object):

    def test_empty_config(self):
        config = Config()
        config.merge(Config())
        compare(config.data, expected={})

    def test_non_empty_config(self):
        config = Config({'foo': 'bar'})
        config.merge(Config({'baz': 'bob'}))
        compare(config.data, {'foo': 'bar', 'baz': 'bob'})

    def test_simple_type(self):
        config = Config()
        with ShouldRaise(type_error(
            "Cannot merge <class 'str'> with <class 'dict'>"
        )):
            config.merge('foo')

    def test_dict_to_dict(self):
        config = Config({'x': 1})
        config.merge({'y': 2})
        compare(config.data, expected={'x': 1, 'y': 2})

    def test_list_to_list(self):
        config = Config([1, 2])
        config.merge([3, 4])
        compare(config.data, expected=[1, 2, 3, 4])

    def test_dict_to_list(self):
        config1 = Config([1, 2])
        config2 = Config({'x': 1})
        with ShouldRaise(type_error(
            "Cannot merge <class 'dict'> with <class 'list'>"
        )):
            config1.merge(config2)

    def test_list_to_dict(self):
        config1 = Config({'x': 1})
        config2 = Config([1, 2])
        with ShouldRaise(type_error(
            "Cannot merge <class 'list'> with <class 'dict'>"
        )):
            config1.merge(config2)

    def test_other_to_dict(self):
        config1 = Config(1)
        config2 = Config(1)
        with ShouldRaise(type_error(
            "Cannot merge <class 'int'> with <class 'int'>"
        )):
            config1.merge(config2)

    def test_nested_working(self):
        config1 = Config(dict(x=1, y=[2, 3], z=dict(a=4, b=5)))
        config2 = Config(dict(w=6, y=[7], z=dict(b=8, c=9)))
        config1.merge(config2)

        compare(config1.data,
                expected=dict(x=1, w=6, y=[2, 3, 7], z=dict(a=4, b=8, c=9)))

    def test_override_type_mapping(self):
        config1 = Config([1, 2])
        config2 = Config([3, 4])
        def zipper(context, source, target):
            return zip(target, source)
        config1.merge(config2, mergers={list: zipper})
        compare(config1.data, expected=[(1, 3), (2, 4)])

    def test_type_returns_new_object(self):
        config1 = Config((1, 2))
        config2 = Config((3, 4))
        def concat(context, source, target):
            return target + source
        config1.merge(config2, mergers={tuple: concat})
        compare(config1.data, expected=(1, 2, 3, 4))

    def test_blank_type_mapping(self):
        config1 = Config({'foo': 'bar'})
        config2 = Config({'baz': 'bob'})
        with ShouldRaise(type_error(
            "Cannot merge <class 'dict'> with <class 'dict'>"
        )):
            config2.merge(config1, mergers={})

    def test_supplement_type_mapping(self):
        config1 = Config({'x': (1, 2)})
        config2 = Config({'x': (3, 4)})
        def concat(context, source, target):
            return target + source
        config1.merge(config2, mergers=default_mergers+{tuple: concat})
        compare(config1.data, expected={'x': (1, 2, 3, 4)})

    def test_mapping_paths(self):
        config = Config({'x': 'old'})
        data = {'foo': 'bar'}
        config.merge(data, mapping={
            source['foo']: target['x']
        })
        compare(config.data, expected={'x': 'bar'})

    def test_mapping_strings(self):
        config = Config({'x': 'old'})
        data = {'foo': 'bar'}
        config.merge(data, mapping={
            'foo': 'x'
        })
        compare(config.data, expected={'x': 'bar'})

    def test_mapping_dotted_strings(self):
        config = Config({'a': {'b': 'old'}})
        data = {'c': {'d': 'new'}}
        config.merge(data, mapping={
            'c.d': 'a.b'
        })
        compare(config.data, expected={'a': {'b': 'new'}})

    def test_mapping_type_conversion(self):
        config = Config({'x': 0})
        data = {'y': '1'}
        config.merge(data, mapping={
            convert(source['y'], int): target['x']
        })
        compare(config.data, expected={'x': 1})

    def test_mapping_extensive_conversation(self):
        config = Config({'a': 0})
        data = {'x': 2, 'y': -1}

        def best(possible):
            return max(possible.values())

        config.merge(data, mapping={
            convert(source, best): target['a']
        })

        compare(config.data, expected={'a': 2})

    def test_mapping_with_merge(self):
        config = Config({'x': {'y': 1}})
        data = {'z': 2}
        config.merge(data, mapping={
            source: target['x'].merge()
        })
        compare(config.data, expected={'x': {'y': 1, 'z': 2}})

    def test_mapping_with_top_level_merge(self):
        config = Config({'x': {'y': 1}})
        data = {'z': 2}
        config.merge(data, mapping={
            source: target.merge()
        })
        compare(config.data, expected={'x': {'y': 1}, 'z': 2})


class TestAddition(object):

    def test_top_level_dict(self):
        config1 = Config({'foo': 'bar'})
        config2 = Config({'baz': 'bob'})
        config3 = config1 + config2
        compare(config1.data, {'foo': 'bar'})
        compare(config2.data, {'baz': 'bob'})
        compare(config3.data, {'foo': 'bar', 'baz': 'bob'})

    def test_top_level_list(self):
        config1 = Config([1, 2])
        config2 = Config([3, 4])
        config3 = config1 + config2
        compare(config1.data, [1, 2])
        compare(config2.data, [3, 4])
        compare(config3.data, [1, 2, 3, 4])

    def test_non_config_rhs(self):
        config = Config({'foo': 'bar'}) + {'baz': 'bob'}
        compare(config.data, {'foo': 'bar', 'baz': 'bob'})

    def test_failure(self):
        with ShouldRaise(type_error(
            "Cannot merge <class 'int'> with <class 'dict'>"
        )):
            Config({'foo': 'bar'}) + 1



from argparse import ArgumentParser

import pytest
from testfixtures import TempDirectory, compare

from configurator import Config
from configurator.mapping import target, convert


@pytest.fixture
def dir():
    with TempDirectory(encoding='ascii') as d:
        yield d


class TestFunctional(object):

    def test_layered(self, dir):
        # defaults
        config = Config({
            'database': {'user': 'foo'},
            'special': False
        })
        # from system file:
        path = dir.write('etc/myapp.json', '{"special": true}')
        config.merge(Config.from_path(path))
        # from user file:
        path = dir.write('home/user/myapp.json', '{"database": {"password": "123"}}')
        config.merge(Config.from_path(path))
        # end result:
        compare(config.database.user, expected='foo')
        compare(config.database.password, expected='123')
        compare(config.special, expected=True)

    def test_defaults_and_argparse(self):
        parser = ArgumentParser()
        parser.add_argument('--first-url')
        parser.add_argument('--attempts', type=int, default=2)
        parser.add_argument('--bool-a', action='store_true')
        parser.add_argument('--bool-b', action='store_false')

        args = parser.parse_args(['--first-url', 'override_url', '--bool-a'])
        compare(args.bool_b, expected=True)

        config = Config({'urls': ['default_url'], 'bool_b': False})

        config.merge(args, {
            'first_url': target['urls'].insert(0),
            'attempts': target['attempts'],
            'bool_a': target['bool_a'],
            'bool_b': target['bool_b'],
        })

        compare(config.urls, expected=['override_url', 'default_url'])
        compare(config.attempts, expected=2)
        compare(config.bool_a, expected=True)
        compare(config.bool_b, expected=True)

    def test_defaults_and_env(self):
        config = Config({'present': 'dp', 'absent': 'da'})

        environ = {'ENV_PRESENT': '1'}

        config.merge(environ, {
            convert('ENV_PRESENT', int): 'present',
            convert('ENV_ABSENT', int): 'absent',
        })

        compare(config.present, expected=1)
        compare(config.absent, expected='da')

    def test_overlay(self, dir):
        pytest.importorskip("yaml")
        path1 = dir.write('etc/myapp.yml', '''
        base: 1
        user: bad
        file: bad
        ''')

        path2 = dir.write('home/.myapp.yml', '''
        user: 2
        file: bad-user
        ''')

        path3 = dir.write('app.yml', '''
        file: 3
        ''')

        config = Config.from_path(path1)
        config.merge(Config.from_path(path2))
        config.merge(Config.from_path(path3))

        compare(config.base, expected=1)
        compare(config.user, expected=2)
        compare(config.file, expected=3)


def test_fake_fs(fs):
    fs.create_file('/foo/bar.yml', contents='foo: 1\n')
    config = Config.from_path('/foo/bar.yml')
    compare(config.foo, expected=1)



from .config import Config
from .merge import default_mergers
from .mapping import source, target, convert, required, if_supplied

__all__ = (
    'Config', 'source', 'target', 'convert', 'required', 'default_mergers', 'if_supplied'
)



from textwrap import dedent

from testfixtures import compare, ShouldRaise

from configurator.node import ConfigNode


class TestInstantiation(object):

    def test_empty(self):
        config = ConfigNode()
        compare(config.data, expected={})

    def test_dict(self):
        config = ConfigNode(dict(x=1))
        compare(config['x'], expected=1)

    def test_list(self):
        config = ConfigNode([1, 2])
        compare(config[0], expected=1)
        compare(config[1], expected=2)
        compare(list(config), expected=[1, 2])

    def test_int(self):
        # not very useful, but documents .data as a public api
        config = ConfigNode(1)
        compare(config.data, 1)


class TestDictAccess(object):

    def test_there_dict(self):
        config = ConfigNode({'foo': 1})
        compare(config['foo'], expected=1)

    def test_there_list(self):
        config = ConfigNode([1])
        compare(config[0], expected=1)

    def test_not_there_dict(self):
        config = ConfigNode({})
        with ShouldRaise(KeyError('foo')):
            config['foo']

    def test_not_there_list(self):
        config = ConfigNode([])
        with ShouldRaise(IndexError('list index out of range')):
            config[0]

    def test_not_there_simple(self):
        config = ConfigNode(1)
        with ShouldRaise(KeyError('foo')):
            config['foo']

    def test_get(self):
        config = ConfigNode({'foo': 1})
        compare(config.get('foo'), expected=1)

    def test_get_default(self):
        config = ConfigNode({})
        compare(config.get('foo', 1), expected=1)

    def test_get_default_default(self):
        config = ConfigNode({})
        compare(config.get('foo'), expected=None)

    def test_get_list_number_key(self):
        config = ConfigNode([1])
        compare(config.get(0), expected=1)

    def test_get_list_string_key(self):
        config = ConfigNode([1])
        compare(config.get('foo'), expected=None)

    def test_simple_value(self):
        config = ConfigNode({})
        compare(config.get('foo', 1), expected=1)

    def test_dict_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_list_value(self):
        config = ConfigNode({'foo': [1]})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_get_wrapped_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config['foo']
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_items_dict(self):
        config = ConfigNode({'foo': 1, 'bar': 2})
        compare(config.items(), expected=[('bar', 2), ('foo', 1)])

    def test_items_list(self):
        config = ConfigNode([])
        expected = "'list' object has no attribute 'items'"
        with ShouldRaise(AttributeError(expected)):
            tuple(config.items())

    def test_items_wrapped_value(self):
        config = ConfigNode({'foo': [1]})
        obj = tuple(config.items())[0][1]
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_set_item(self):
        config = ConfigNode()
        with ShouldRaise(TypeError):
            config['foo'] = 1
        compare(config.data, expected={})


class TestAttributeAccess(object):

    def test_there(self):
        config = ConfigNode({'foo': 1})
        compare(config.foo, expected=1)

    def test_not_there(self):
        config = ConfigNode({})
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_simple_value(self):
        config = ConfigNode({'foo': 1})
        compare(config.foo, expected=1)

    def test_dict_value(self):
        config = ConfigNode({'foo': {'bar': 1}})
        obj = config.foo
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected={'bar': 1})

    def test_list_value(self):
        config = ConfigNode({'foo': [1]})
        obj = config.foo
        assert isinstance(obj, ConfigNode)
        compare(obj.data, expected=[1])

    def test_name_clash_with_real_attrs(self):
        # method, data, __method
        config = ConfigNode({'get': 1})
        compare(config.get('get'), expected=1)

    def test_list_source(self):
        config = ConfigNode([1])
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_simple_source(self):
        config = ConfigNode(1)
        with ShouldRaise(AttributeError('foo')):
            config.foo

    def test_set_attr(self):
        config = ConfigNode()
        with ShouldRaise(AttributeError):
            config.foo = 1
        compare(config.data, expected={})


class TestOtherFunctionality(object):

    def test_iterate_over_list_of_dicts(self):
        node = ConfigNode([{'x': 1}])
        compare(tuple(node)[0], expected=ConfigNode({'x': 2}))

    def test_repr(self):
        node = ConfigNode({'some long key': 'some\nvalue',
                            'another long key': 2,
                            'yet another long key': 3})
        compare(repr(node), expected=dedent("""\
            configurator.node.ConfigNode(
            {'another long key': 2,
             'some long key': 'some\\nvalue',
             'yet another long key': 3}
            )"""))



from .path import Path, parse_text, ConvertOp, RequiredOp, NotPresent, IfSuppliedOp


def load(data, path):
    path = parse_text(path)
    for op in path.ops:
        if isinstance(data, NotPresent):
            op.not_present(data)
        else:
            data = op.get(data)
    return data


def convert(source, callable_):
    """
    A :doc:`mapping <mapping>` operation that indicates the source value
    should be converted by calling ``callable_`` with the original value
    and then using that result from that point in the mapping operation
    onwards.
    """
    source = parse_text(source)
    return source._extend(ConvertOp(callable_))


def required(source):
    """
    A :doc:`mapping <mapping>` operation that indicates the source value
    is required. If it is not present, the exception that occurred when
    trying to obtain it will be raised.
    """
    source = parse_text(source)
    return source._extend(RequiredOp())


def if_supplied(source):
    """
    A :doc:`mapping <mapping>` operation that indicates the source value
    should be treated as not present if its value matches Python's definition
    of "false". This includes empty strings, zero, ``False`` and ``None``.
    """
    source = parse_text(source)
    return source._extend(IfSuppliedOp())


def store(data, path, value, merge_context=None):
    path = parse_text(path)
    if not path.ops:
        raise TypeError('Cannot store at root')
    stack = [data]
    for op in path.ops[:-1]:
        stack.append(op.ensure(stack[-1]))
    if not isinstance(value, NotPresent):
        data = path.ops[-1].set(stack[-1], value, merge_context)
        if data is not None:
            # uh oh, we have to replace the upstream object:
            if len(stack) < 2:
                stack[0] = data
            else:
                path.ops[-2].set(stack[-2], data, merge_context)
    return stack[0]


source = Path('source')
target = Path('target')



import linecache
import sys
import tokenize
from doctest import REPORT_NDIFF, ELLIPSIS

import pytest
from pyfakefs.fake_filesystem_unittest import Patcher
from sybil import Sybil
from sybil.parsers.doctest import DocTestParser, FIX_BYTE_UNICODE_REPR
from sybil.parsers.codeblock import CodeBlockParser
from testfixtures import Replacer, TempDirectory
from testfixtures.sybil import FileParser

@pytest.fixture(scope='module')
def fs_state():
    pytest.importorskip("yaml")
    if sys.version_info < (3, 6):
        pytest.skip('docs are py3 only')
    patcher = Patcher(additional_skip_names=['expanduser'])
    patcher.setUp()
    patcher.pause()
    yield patcher
    patcher.tearDown()


@pytest.fixture()
def fs(fs_state):
    # We need our own one to have state across sections in documents.
    fs_state.resume()
    linecache.open = fs_state.original_open
    tokenize._builtin_open = fs_state.original_open
    yield fs_state.fs
    fs_state.pause()


@pytest.fixture()
def tempdir(fs):
    with TempDirectory(path='/') as d:
        yield d


@pytest.fixture(scope='module')
def replace():
    with Replacer() as r:
        yield r.replace


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(optionflags=REPORT_NDIFF|ELLIPSIS|FIX_BYTE_UNICODE_REPR),
        CodeBlockParser(['print_function']),
        FileParser('tempdir')
    ],
    pattern='*.rst',
    fixtures=['fs', 'replace', 'tempdir'],
).pytest()



class NotPresent(Exception): pass


class Op(object):

    name = 'op'

    def get(self, data):
        raise TypeError('Cannot use %s() in source' % self.name)

    def ensure(self, *args):
        raise TypeError('Cannot use %s() as target' % self.name)

    set = ensure

    def not_present(self, data):
        return data


class ItemOp(Op):

    def __init__(self, text):
        self.text = text

    def get(self, data):
        try:
            return data[self.text]
        except (KeyError, IndexError):
            return NotPresent(self.text)

    def ensure(self, data):
        try:
            return data[self.text]
        except KeyError:
            data[self.text] = value = {}
            return value

    def set(self, data, value, _):
        data[self.text] = value

    def str(self, base):
        return '{}[{!r}]'.format(base, self.text)


class AttrOp(Op):

    def __init__(self, text):
        self.text = text

    def get(self, data):
        return getattr(data, self.text, NotPresent(self.text))

    def ensure(self, data):
        return getattr(data, self.text)

    def set(self, data, value, _):
        setattr(data, self.text, value)

    def str(self, base):
        return '{}.{}'.format(base, self.text)


class TextOp(Op):

    def __init__(self, text):
        self.text = text

    def get(self, data):
        getitem = getattr(data, '__getitem__', None)
        if getitem is None:
            return getattr(data, self.text, NotPresent(self.text))
        else:
            try:
                return getitem(self.text)
            except KeyError:
                return NotPresent(self.text)

    def ensure(self, data):
        getitem = getattr(data, '__getitem__', None)
        if getitem is None:
            return getattr(data, self.text)
        else:
            try:
                return getitem(self.text)
            except KeyError:
                data[self.text] = value = {}
                return value

    def set(self, data, value, _):
        setitem = getattr(data, '__setitem__', None)
        if setitem is None:
            return setattr(data, self.text, value)
        else:
            return setitem(self.text, value)

    def str(self, base):
        if base:
            return base+'.'+self.text
        else:
            return self.text


class ConvertOp(Op):

    name = 'convert'

    def __init__(self, callable_):
        self.callable = callable_

    def get(self, data):
        return self.callable(data)

    def str(self, base):
        callable_str = getattr(self.callable, '__name__', repr(self.callable))
        return '{}({}, {})'.format(self.name, base, callable_str)


class RequiredOp(Op):

    name = 'required'

    def get(self, data):
        return data

    def not_present(self, data):
        raise data

    def str(self, base):
        return '{}({})'.format(self.name, base)


class IfSuppliedOp(Op):

    name = 'if_supplied'

    def get(self, data):
        return data or NotPresent(data)

    def str(self, base):
        return '{}({})'.format(self.name, base)


class InsertOp(Op):

    name = 'insert'

    def __init__(self, index):
        self.index = index

    def ensure(self, data):
        value = {}
        data.insert(self.index, value)
        return value

    def set(self, data, value, _):
        data.insert(self.index, value)

    def str(self, base):
        return '{}.{}({!r})'.format(base, self.name, self.index)


class AppendOp(Op):

    name = 'append'

    def ensure(self, data):
        value = {}
        data.append(value)
        return value

    def set(self, data, value, _):
        data.append(value)

    def str(self, base):
        return '{}.{}()'.format(base, self.name)


class MergeOp(Op):

    name = 'merge'

    def ensure(self, data):
        raise TypeError('merge() must be final operation')

    def set(self, data, value, context):
        return context.merge(data, value)

    def str(self, base):
        return '{}.{}()'.format(base, self.name)


class Path(object):
    """
    A generative object used for constructing source or target mappings.
    See :doc:`mapping` for details.
    """

    def __init__(self, name, *ops):
        self.name = name
        self.ops = ops

    def _extend(self, op):
        return type(self)(self.name, *(self.ops + (op,)))

    def __getitem__(self, name):
        """
        Indicate that the source or target should be traversed by item access.
        """
        return self._extend(ItemOp(name))

    def __getattr__(self, name):
        """
        Indicate that the source or target should be traversed by attribute access.
        """
        return self._extend(AttrOp(name))

    def insert(self, index):
        """
        Indicate that a target should be mapped by inserting at the specified
        index.
        """
        return self._extend(InsertOp(index))

    def append(self):
        """
        Indicate that a target should be mapped by appending.
        """
        return self._extend(AppendOp())

    def merge(self):
        """
        Indicate that a target should be mapped by merging.
        """
        return self._extend(MergeOp())

    def __str__(self):
        str = self.name
        for op in self.ops:
            str = op.str(str)
        return str

    def __repr__(self):
        return 'Path:{}'.format(str(self))


def parse_text(segment):
    if isinstance(segment, str):
        segment = Path('', *(TextOp(part) for part in segment.split('.')))
    return segment



from testfixtures import compare

from configurator.mapping import source, target, required, convert
from configurator.path import parse_text


class TestPaths(object):

    def test_repr(self):
        compare(repr(source), expected="Path:source")
        compare(repr(target), expected="Path:target")

    def test_str(self):
        compare(str(source), expected="source")
        compare(str(target), expected="target")

    def test_repr_nested(self):
        compare(repr(
            required(convert(source['foo'].y, int)).insert(0).append().merge()
        ), expected=(
            "Path:required(convert(source['foo'].y, int)).insert(0).append().merge()"
        ))

    def test_str_nested(self):
        compare(str(
            required(convert(source['foo'].y, int)).insert(0).append().merge()
        ), expected=(
            "required(convert(source['foo'].y, int)).insert(0).append().merge()"
        ))

    def test_text_op(self):
        compare(str(parse_text('x.y.z')), expected='x.y.z')

    def test_convert_no_name(self):
        o = object()
        compare(str(convert(source, o)), expected=(
            "convert(source, {!r})".format(o)
        ))



from argparse import Namespace
from testfixtures import compare, ShouldRaise

from configurator.mapping import source, load, convert, store, target, required, if_supplied
from configurator.merge import MergeContext
from configurator.path import NotPresent


class TestSource(object):

    def test_root(self):
        data = {'foo'}
        compare(load(data, source), expected=data)

    def test_getitem(self):
        data = {'foo': 'bar'}
        compare(load(data, source['foo']), expected='bar')

    def test_index(self):
        data = ['a', 'b']
        compare(load(data, source[1]), expected='b')

    def test_attr(self):
        data = Namespace(x=1)
        compare(load(data, source.x), expected=1)

    def test_required_and_present(self):
        data = ['a', 'b']
        compare(load(data, required(source[1])), expected='b')

    def test_getitem_not_present(self):
        data = {}
        with ShouldRaise(NotPresent('foo')):
            load(data, required(source['foo']))

    def test_getitem_nested_not_present(self):
        data = {}
        with ShouldRaise(NotPresent('foo')):
            load(data, required(source['foo']['bar']))

    def test_index_not_present(self):
        data = ['a']
        with ShouldRaise(NotPresent(1)):
            load(data, required(source[1]))

    def test_attr_not_present(self):
        data = Namespace()
        with ShouldRaise(NotPresent('x')):
           load(data, required(source.x))

    def test_attr_nested_not_present(self):
        data = Namespace()
        with ShouldRaise(NotPresent('x')):
           load(data, required(source.x.y))

    def test_getitem_not_present_okay(self):
        data = {}
        compare(load(data, source['foo']), expected=NotPresent('foo'))

    def test_index_not_present_okay(self):
        data = ['a']
        compare(load(data, source[1]), expected=NotPresent(1))

    def test_attr_not_present_okay(self):
        data = Namespace()
        compare(load(data, source.x), expected=NotPresent('x'))

    def test_nested(self):
        data = {'foo': ['a', 'b', Namespace(x=1)]}
        compare(load(data, source['foo'][2].x), expected=1)

    def test_nested_missing_okay(self):
        data = {'foo': []}
        compare(load(data, source['foo'][2].x),
                expected=NotPresent(2))

    def test_string_item(self):
        data = {'foo': 'bar'}
        compare(load(data, 'foo'), expected='bar')

    def test_string_attr(self):
        data = Namespace(foo='bar')
        compare(load(data, 'foo'), expected='bar')

    def test_string_item_not_present(self):
        data = {}
        compare(load(data, 'foo'), expected=NotPresent('foo'))

    def test_string_attr_not_present(self):
        data = Namespace()
        compare(load(data, 'foo'), expected=NotPresent('foo'))

    def test_string_dotted(self):
        data = {'foo': Namespace(x=1)}
        compare(load(data, 'foo.x'), expected=1)

    def test_string_item_not_present_required(self):
        data = {}
        with ShouldRaise(NotPresent('foo')):
            load(data, required('foo'))

    def test_convert(self):
        data = Namespace(x='1')
        compare(load(data, convert(source.x, int)), expected=1)

    def test_convert_string(self):
        data = Namespace(x='1')
        compare(load(data, convert('x', int)), expected=1)

    def test_convert_not_present(self):
        data = {}
        compare(load(data, convert('x', int)), expected=NotPresent('x'))

    def test_insert(self):
        with ShouldRaise(TypeError('Cannot use insert() in source')):
            load(None, source.insert(0))

    def test_append(self):
        with ShouldRaise(TypeError('Cannot use append() in source')):
            load(None, source.append())

    def test_merge(self):
        with ShouldRaise(TypeError('Cannot use merge() in source')):
            load(None, source.merge())

    def test_if_supplied_truthy(self):
        data = Namespace(x='1')
        compare(load(data, if_supplied(source.x)), expected='1')

    def test_if_supplied_falsy(self):
        data = Namespace(x=None)
        compare(load(data, if_supplied(source.x)), expected=NotPresent(None))

    def test_if_supplied_string(self):
        data = Namespace(x='1')
        compare(load(data, if_supplied('x')), expected='1')

    def test_if_supplied_required_falsy(self):
        data = Namespace(x=None)
        with ShouldRaise(NotPresent(None)):
            load(data, required(if_supplied(source.x)))

    def test_if_supplied_str(self):
        compare(str(if_supplied(source.x)), expected='if_supplied(source.x)')


class TestTarget(object):

    def test_root(self):
        data = {'foo'}
        with ShouldRaise(TypeError('Cannot store at root')):
            store(data, target, 'foo')
        compare(data, expected={'foo'})

    def test_getitem(self):
        data = {'foo': 'bar'}
        store(data, target['foo'], 'baz')
        compare(data, expected={'foo': 'baz'})

    def test_getitem_not_present(self):
        data = {'foo': 'bar'}
        store(data, target['foo'], NotPresent('foo'))
        compare(data, expected={'foo': 'bar'})

    def test_index(self):
        data = ['a', 'b']
        store(data, target[1], 'c')
        compare(data, expected=['a', 'c'])

    def test_index_not_present(self):
        data = ['a', 'b']
        store(data, target[1], NotPresent(1))
        compare(data, expected=['a', 'b'])

    def test_append(self):
        data = ['a', 'b']
        store(data, target.append(), 'c')
        compare(data, expected=['a', 'b', 'c'])

    def test_append_nested(self):
        data = []
        store(data, target.append()['a'], 'b')
        compare(data, expected=[{'a': 'b'}])

    def test_append_not_present(self):
        data = []
        store(data, target.append()['a'], NotPresent('foo'))
        compare(data, expected=[{}])

    def test_insert(self):
        data = ['a', 'b']
        store(data, target.insert(1), 'c')
        compare(data, expected=['a', 'c', 'b'])

    def test_insert_nested(self):
        data = []
        store(data, target.insert(0)['a'], 'b')
        compare(data, expected=[{'a': 'b'}])

    def test_insert_not_present(self):
        data = []
        store(data, target.insert(0)['a'], NotPresent('foo'))
        compare(data, expected=[{}])

    def test_attr(self):
        data = Namespace(x=1)
        store(data, target.x, 2)
        compare(data.x, 2)

    def test_attr_not_present(self):
        data = Namespace(x=1)
        store(data, target.x, NotPresent('foo'))
        compare(data.x, 1)

    def test_nested(self):
        data = {'foo': ['a', 'b', Namespace(x=1)]}
        store(data, target['foo'][2].x, 2)
        compare(data['foo'][2].x, expected=2)

    def test_string_item(self):
        data = {'foo': 'bar'}
        store(data, 'foo', 'baz')
        compare(data, expected={'foo': 'baz'})

    def test_string_attr(self):
        data = Namespace(foo='bar')
        store(data, target.x, 2)
        compare(data.x, 2)

    def test_string_dotted(self):
        data = {'foo': Namespace(x=1)}
        store(data, 'foo.x', 2)
        compare(data['foo'].x, expected=2)

    def test_create_nested_dicts(self):
        # created needed dicts
        data = {}
        store(data, target['x']['y'], 2)
        compare(data, expected={'x': {'y': 2}})

    def test_create_nested_attrs(self):
        # exception
        data = {}
        with ShouldRaise(AttributeError("'dict' object has no attribute 'x'")):
            store(data, target.x.y, 2)

    def test_create_from_dotted_string(self):
        data = {}
        store(data, 'x.y', 2)
        compare(data, expected={'x': {'y': 2}})

    def test_create_nested_attrs_from_dotted_string(self):
        # exception
        data = Namespace()
        with ShouldRaise(AttributeError(
            "'Namespace' object has no attribute 'x'"
        )):
            store(data, 'x.y', 2)

    def test_set_on_convert(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use convert() as target')):
            store(data, convert(target, int), 'y')

    def test_ensure_on_convert(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use convert() as target')):
            store(data, convert(target, int).x, 'y')

    def test_set_on_required(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use required() as target')):
            store(data, required(target), 'y')

    def test_ensure_on_required(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use required() as target')):
            store(data, required(target).x, 'y')

    def test_set_on_supplied(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use if_supplied() as target')):
            store(data, if_supplied(target), 'y')

    def test_ensure_on_supplied(self):
        data = '1'
        with ShouldRaise(TypeError('Cannot use if_supplied() as target')):
            store(data, if_supplied(target).x, 'y')

    def test_merge(self):
        data = {'x': 1}
        data = store(data, target.merge(), {'y': 2}, MergeContext())
        compare(data, expected={'x': 1, 'y': 2})

    def test_merge_nested(self):
        data = {'x': {'y': 2}}
        store(data, target['x'].merge(), {'z': 1}, MergeContext())
        compare(data, expected={'x': {'y': 2, 'z': 1}})

    def test_merge_not_present(self):
        data = {'x': {'y': 2}}
        store(data, target['x'].merge(), NotPresent('foo'), MergeContext())
        compare(data, expected={'x': {'y': 2}})

    def test_ensure_on_merge(self):
        data = {}
        with ShouldRaise(TypeError('merge() must be final operation')):
            store(data, target.merge().x, 'y', MergeContext())



import os
from copy import deepcopy
from io import open, StringIO
from os.path import exists, expanduser

from .mapping import load, store, target, convert, if_supplied
from .merge import MergeContext
from .node import ConfigNode
from .parsers import Parsers
from .path import parse_text


class Config(ConfigNode):
    """
    The root of the configuration store.
    """

    __slots__ = ('data', '_previous')

    parsers = Parsers.from_entrypoints()

    def __init__(self, data=None):
        super(Config, self).__init__(data)
        self._previous = []

    @classmethod
    def from_text(cls, text, parser, encoding='ascii'):
        """
        Construct a :class:`Config` from the provided ``text`` using the specified
        :doc:`parser <parsers>`. If ``text`` is provided as :class:`bytes`, then
        the ``encoding`` specified will be used to decode it.
        """
        if isinstance(text, bytes):
            text = text.decode(encoding)
        return cls.from_stream(StringIO(text), parser)

    @classmethod
    def from_stream(cls, stream, parser=None):
        """
        Construct a :class:`Config` from a stream such as a :func:`file <open>`.
        If the stream does not have a ``name`` attribute from which the correct
        parser can be guessed, :doc:`parser <parsers>` must be specified.
        If ``text`` is provided as :class:`bytes`, then the ``encoding``
        specified will be used to decode it.
        """
        if parser is None:
            name = getattr(stream, 'name', None)
            if name is not None:
                try:
                    _, parser = name.rsplit('.', 1)
                except ValueError:
                    pass
        if not callable(parser):
            parser = cls.parsers.get(parser)
        return cls(parser(stream))

    @classmethod
    def from_path(cls, path, parser=None, encoding=None, optional=False):
        """
        Construct a :class:`Config` from file specified as a path. This path
        with have ``~`` expanded. If specified, ``encoding`` will be used to
        decode the content of the file. An explicit :doc:`parser <parsers>`
        can be specified, if necessary, but the correct one will usually be
        guessed from the file extension.

        If ``optional`` is ``True``, then an empty :class:`Config` will be
        returned if the file does not exist.
        """
        full_path = expanduser(path)
        if optional and not exists(full_path):
            return cls()
        with open(full_path, encoding=encoding) as stream:
            return cls.from_stream(stream, parser)

    @classmethod
    def from_env(cls, prefix, types=None):
        """
        Construct a :class:`Config` from :data:`os.environ` entries
        that matches the specified ``prefix``.
        ``prefix`` can either be a simple string prefix
        or a :class:`dict` mapping string prefixes to the
        :doc:`target paths <mapping>` at which they will be stored.
        ``types`` is an optional :class:`dict` mapping string suffixes
        to callables used to convert matching environment values to the
        correct type.
        """
        if not isinstance(prefix, dict):
            prefixes = {prefix: target}
        else:
            prefixes = prefix
        mapping = {}
        for key, value in os.environ.items():
            for prefix, prefix_target in prefixes.items():
                if key.startswith(prefix):
                    prefix_source = if_supplied(key)
                    prefix_target = parse_text(prefix_target)[key[len(prefix):].lower()]
                    if types is not None:
                        for suffix, type_ in types.items():
                            if key.endswith(suffix):
                                prefix_source = convert(prefix_source, type_)
                    mapping[prefix_source] = prefix_target

        config = cls()
        config.merge(os.environ, mapping)
        return config

    def merge(self, source, mapping=None, mergers=None):
        """
        Modify this :class:`Config` by merging the provided ``source`` into
        it using any ``mapping`` or ``mergers`` provided.

        See :doc:`mapping` for more detail.
        """
        if isinstance(source, Config):
            source = source.data
        context = MergeContext(mergers)
        if mapping is None:
            self.data = context.merge(source, self.data)
        else:
            for source_path, target_path in mapping.items():
                value = load(source, source_path)
                self.data = store(self.data, target_path, value, context)

    def __add__(self, other):
        """
        Configuration stores can be added together.
        This will result in a new :class:`Config` object being returned that
        is created by merging the two original configs.
        """
        result = Config(type(self.data)())
        result.merge(self)
        result.merge(other)
        return result

    def push(self, config=None, empty=False):
        """
        Push the provided ``config`` onto this instance, replacing the data
        of this :class:`Config`.

        If ``empty`` is ``False``, this is done by merging the existing
        contents with the provided ``config``, giving precedence to the
        ``config`` passed in.

        If ``empty`` is ``True``, then the provided ``config`` is used to
        entirely replace the data of this :class:`Config`.

        ``config`` may either be a :class:`Config` instance or anything
        that would be passed to the :class:`Config` constructor.

        This method returns a context manager that, when its context is left,
        restores the configuration data used to whatever was in place
        before :meth:`push` was called, regardless of any further :meth:`push`
        or meth:`merge` calls, or other modifications to :attr:`data` on this
        :class:`Config` object.
        """
        if empty:
            base = Config()
        else:
            base = Config(deepcopy(self.data))
        if not isinstance(config, Config):
            config = Config(config)
        self._previous.append(self.data)
        context = PushContext(self, self.data)
        self.data = (base + config).data
        return context

    def pop(self):
        """
        Pop off the top-most data that was last :meth:`pushed <push>`
        on to this :class:`Config`.
        """
        self.data = self._previous.pop()


class PushContext(object):

    def __init__(self, config, data):
        self.config = config
        self.data = data

    def __enter__(self):
        return self.config

    def __exit__(self, exc_type, exc_val, exc_tb):
        while self.config.data is not self.data:
            self.config.pop()
